/**********************************************************************************************
 *****************      FINAL PROJECT IN NEURAL NETWORKS -- CS 5388     ***********************
 *****************                 by: Karan M. Gupta                   ***********************
 **********************************************************************************************
 *****************           2-Layer Back Propagation Network           ***********************
 *****************                  Lecture 7, Pg 24                    ***********************
 **********************************************************************************************/

void BPN()
{

  cout << endl << "************           2-Layer Back Propagation Network            ************";
  cout << endl << "************                   Lecture 7, Pg 24                    ************" << endl;

  const int P = 4;                      // no. of training patterns: (z,d) pairs
  const int X = 3;                      // no. of elements in augmented input vector x
  const int I = 3;                      // no. of vectors in v vector
  const int J = 3;                      // no. of elements in y, w vectors
  const double Emax = 0.01 ;                // maximum error

  /**** Results: **********************************************************
   ****	Net seems moderately trained after 1451 cycles, 
   **** using a value of 0.1 for the Maximum Error allowed.
   **** Decreasing the value of Emax to 0.01 improves the performance of
   **** the NN however at the expense of time, as the NN now takes 3096
   **** cycles to learn optimum weight vector values.
   ****
   **** After the net has trained itself, the program runs a test to check
   **** whether the NN does actually output the expected correct value for
   **** the XOR truth table or not. For this test the NN uses the learned
   **** values of V and W weight vectors and calculates values for the
   **** output O as per the input vector X.
   ************************************************************************/
  typedef CWTSVector<X> VectorX;
  typedef CWTSVector<J> VectorY;

  double threshold = -1.0;
  double nu = 0.1;
  double lambda = 1;
  int cycleno;
  int p, i, j, k;
  double e, o, netY, netO, delO;

  VectorX x[P];
  x[0][0] = 0; x[0][1] = 0; x[0][2] = threshold; 
  x[1][0] = 0; x[1][1] = 1; x[1][2] = threshold; 
  x[2][0] = 1; x[2][1] = 0; x[2][2] = threshold; 
  x[3][0] = 1; x[3][1] = 1; x[3][2] = threshold; 

  VectorX v[I];
  v[0][0] = 0.5; v[0][1] = 0.4; v[0][2] = 0.8;
  v[1][0] = 0.9; v[1][1] = 1.0; v[1][2] = -0.1;
  v[2][0] = 0.0; v[2][1] = 0.0; v[2][2] = 0.0;  // dummy weights

  VectorY y;
  y.Fill(threshold);
  VectorY delY;
  delY.Fill(0);

  VectorY w;
  w[0] = -1.2; w[1] = 1.1; w[2] = 0.3;

  double d[] = {0, 1, 1, 0};

  char choice, c;
  cout << endl << "No. of Input Patterns P (" << P << "): "; 
  cout << endl << "Threshold (" << threshold << "): ";
  cout << endl << "nu (" << nu << "): ";
  cout << endl << "lambda (" << lambda << "): ";
  for (i = 0; i < P; i++)
    {
      cout << endl << "Input Vector X" << i+1; 
      iprint(": ", x[i]);
    }
  for (i = 0; i < I; i++)
    {
      cout << endl << "1st Weight Vector V" << i+1 << ": "; 
      iprint(v[i]);
    }
  cout << endl << "2nd Weight Vector W: "; iprint(w);

  cout << endl << "Do you want to use the default values? (y/n): ";
  cin >> choice;
  if ((choice == 'n') || (choice == 'N'))
    {
      cout << endl << "Preparing to accept values from the user...";
      cout << endl << "The original values are shown in parenthesis." << endl;
      cout << endl << "Threshold (" << threshold << "): "; cin >> threshold;
      cout << "nu (" << nu << "): "; cin >> nu;
      cout << "lambda (" << lambda << "): "; cin >> lambda;
      for (i = 0; i < P; i++)
	for (j = 0; j < X; j++)
	  {
	    cout << "Enter value for X" << i+1 << " " << j << " (" << x[i][j] << "): ";
	    cin >> x[i][j];
	  }
      for (i = 0; i < I; i++)
	for (j = 0; j < X; j++)
	  {
	    cout << "Enter value for V" << i+1 << " " << j << " (" << v[i][j] << "): ";
	    cin >> v[i][j];
	  }
      for (j = 0; j < J; j++)
	{
	  cout << "Enter value for W" << j+1 << " (" << w[j] << "): ";
	  cin >> w[j];
	}
    }

  e = Emax + 1;
  cycleno = 0;  
  while(e > Emax)
    {
      e = 0;
      cycleno++;
      // for every pattern
      cout << endl << endl << "*** LEARNING CYCLE NO. " << cycleno << " ***" << endl;
      for (p = 0; p < P; p++)
	{
	  cout << endl << "Working with pattern X[" << p+1 << "]: "; iprint(x[p], "\n");
	  // for every perceptron in 1st layer
	  for (i = 0; i < I; i++)
	    {
	      netY = v[i] * x[p];
	      y[i] = unipolar(netY);         // calculate y, output from 1st layer
	    }
	  y[I-1] = threshold;
	  // calculate o, output from 2nd layer
	  cout << "y; "; iprint(y, "\n");
	  cout << "  w: "; iprint(w, "\n");
	  netO = w * y;
	  o = unipolar(netO);
	  cout << " o = " << o << "  d[" << p+1 << "] = " << d[p];
	  // calculate error in output
	  e = 0.5 * pow((d[p] - o), 2);
	  cout << "  e = " << e;
	  delO = (d[p] - o) * o * (1 - o);
	  cout << "  delO = " << delO << endl;
	  // for every perceptron in 2nd layer
	  for (j = 0; j < J; j++)
	    {
	      delY[j] = y[j] * w[j] * (1 - y[j]) * delO;
	      w[j] = w[j] + y[j] * nu * delO;	    
	    }
	  delY[J-1] = 0;
	  cout << "  w = "; iprint(w, "\n");
	  // propagate weights
	  for (i = 0; i < I; i++)
	    {
	      for (k = 0; k < X; k++)
		v[i][k] = v[i][k] + x[p][k] * nu * delY[i];
	      cout << "v[" << i+1 << "]: "; iprint(v[i], "  ");
	    }
	}
    }
  
  // Training terminated. Display the values.
  cout << endl << endl << "Training terminated: further weight adjustment unrequired." << endl;
  cout << "Total cycles: " << cycleno << "  e: " << e << endl;
  for (i = 0; i < I; i++)
    {
      cout << "v[" << i+1 << "]: "; 
      iprint(v[i], "\n");
    }
  cout << "w: "; iprint(w); cout << endl;

  // Test the Neural Network to see if it produces the correct answers.
  cout << endl << "***********************       TESTING      ***************************";
  //cout << endl << "    X1   |     X2    |       O       |  Final O";
  cout << endl << "   X1\t|\tX2\t|\t    O   \t|    Final O";
  cout << endl << "----------------------------------------------------------------------";
  for (p = 0; p < P; p++)
    {
      //cout << endl << "## Input X[" << p+1 << "]: "; iprint(x[p], "\n");
      for (i = 0; i < I; i++)
	{
	  netY = v[i] * x[p];
	  y[i] = unipolar(netY);
	}
      y[I-1] = threshold;
      netO = w * y;
      o = unipolar(netO);
      cout << endl << "   " << x[p][0] << "\t|\t" << x[p][1] << "\t|\t" << o << "\t|\t" << rint(o);
    }
  cout << endl << "----------------------------------------------------------------------" << endl;
}
