/**********************************************************************************************
 *****************      FINAL PROJECT IN NEURAL NETWORKS -- CS 5388     ***********************
 *****************                 by: Karan M. Gupta                   ***********************
 **********************************************************************************************
 *****************             Kohonen Competitive Learning             ***********************
 *****************                  Lecture 11, Pg 18                   ***********************
 **********************************************************************************************/

void KCL()
{

  cout << endl << "************              Kohonen Competitive Learning             ************";
  cout << endl << "************                   Lecture 11, Pg 18                   ************" << endl;

  const int P = 5;                             // no. of input vectors
  const int n = 2;                             // no. of elements in each input vector
  const int noneurons = 2;                     // no. of neurons

  typedef CWTSVector<n> VectorX;

  /**** Results: *************************************************************************
   **** This implementation of the KCL algorithm uses the example given in the notes.
   **** The user may input different values in the vectors to test the algorithm.
   ***************************************************************************************/

  VectorX x[P];                               // Array of input vectors, i.e. patterns
  x[0][0] = 0.8000; x[0][1] =  0.6000;
  x[1][0] = 0.1736; x[1][1] = -0.9848;
  x[2][0] = 0.7070; x[2][1] =  0.7070;
  x[3][0] = 0.3420; x[3][1] = -0.9397;
  x[4][0] = 0.6000; x[4][1] =  0.8000;

  VectorX w[noneurons];                       // Initialize the weight vector, W
  w[0][0] =  1; w[0][1] = 0;
  w[1][0] = -1; w[1][1] = 0;
      
  double alpha = 1;

  int p, cycleno, i, j;
  double d, minD, dtemp;
  double wtemp;
  int minDneuron;                             // which neuron has the minimum "D"?

  VectorX wOld[noneurons];                      // storing old weight values. will use this to stop training.
  for (i = 0; i < noneurons; i++)
    wOld[i] = w[i];

  char choice, c;

  cout << endl << "No. of Input Patterns P (" << P << "): "; 
  cout << endl << "No. of Neurons (" << noneurons << "): "; 
  cout << endl << "alpha (" << alpha*0.5 << "): ";
  for (i = 0; i < P; i++)
    { cout << endl << "Input Vector X" << i+1; iprint(": ", x[i]);  }
  cout << endl << "The initial weight vector is: " << endl;
  for (i = 0; i < noneurons; i++)
    { cout << "w" << i+1 << ": "; iprint(w[i], "\n"); }
  
  cout << endl << "Do you want to use the default values? (y/n): ";
  cin >> choice;
  if ((choice == 'n') || (choice == 'N'))
    {
      cout << endl << "Preparing to accept values from the user...";
      cout << endl << "The original values are shown in parenthesis." << endl;
      cout << endl << "alpha (" << alpha*0.5 << "): "; cin >> alpha;
      for (i = 0; i < P; i++)
	for (j = 0; j < n; j++)
	  {
	    cout << "Enter value for X" << i+1 << " " << j << " (" << x[i][j] << "): ";
	    cin >> x[i][j];
	  }
      for (i = 0; i < noneurons; i++)
	for (j = 0; j < n; j++)
	  {
	    cout << "Enter value for w" << i+1 << " " << j << " (" << w[i][j] << "): ";
	    cin >> w[i][j];
	  }
    }

  cycleno = 0;
  while(1)
    {
      cycleno++;
      alpha = 0.5 * alpha;
      for (p = 0; p < P; p++)
	{
	  cout << endl << "Input Vector X" << p+1 << ": "; iprint(x[p]);
	  minD = 0;
	  minDneuron = 10;
	  for (i = 0; i < n; i++)
	    minD += pow((x[p][i] - w[0][i]), 2);  // giving minD a large value
	  minD = pow(minD, 0.5);
	  //minD = 10000;
	  for (i = 0; i < noneurons; i++)
	    {
	      dtemp = 0;
	      for (j = 0; j < n; j++)
		dtemp += pow((x[p][j] - w[i][j]), 2);
	      d = pow(dtemp, 0.5);
	      cout << endl << "d" << i+1 << " = " << d;
	      if (d <= minD)
		{
		  minD = d;
		  minDneuron = i;
		  cout << endl << "minD = " << minD << "  minDneuron = " << minDneuron+1;
		}
	    }
	  cout << endl << "The winning neuron is " << minDneuron+1 << " with a value of: " << minD;
	  w[minDneuron] = ( w[minDneuron] + ((x[p] - w[minDneuron]) * alpha) );         
	  // normalizing
	  wtemp = 0;
	  for (i = 0; i < n; i++)
	    wtemp += w[minDneuron][i] * w[minDneuron][i];
	  wtemp = pow(wtemp, 0.5);
	  w[minDneuron] = w[minDneuron] / wtemp;   // end normalizing
	  cout << endl << "The weight vector is: " << endl;
	  for (i = 0; i < noneurons; i++)
	    {
	      cout << "w" << i+1 << ": ";
	      iprint(w[i], "\n");
	    }
	}
      i = 0;
      while(w[i] == wOld[i])
	i++;
      for(j = 0; j < noneurons; j++)
	wOld[i] = w[i];
      if (i == noneurons)
	break;
      /*cout << endl << "Train more? (y/n): ";
      cin >> choice;
      if ((choice == 'n') || (choice == 'N'))
      break;*/
      
    }
  cout << endl << "Training terminated.";
  cout << endl << "The final weight vector is: " << endl;
  for (i = 0; i < noneurons; i++)
    {
      cout << "w" << i+1 << ": ";
      iprint(w[i], "\n");
    }
  cout << "Total no. of cycles: " << cycleno << endl;
}
