/**********************************************************************************************
 *****************      FINAL PROJECT IN NEURAL NETWORKS -- CS 5388     ***********************
 *****************                 by: Karan M. Gupta                   ***********************
 **********************************************************************************************
 *****************  Multiclass Continous Perceptron Training Algorithm  ***********************
 *****************                  Lecture 7, Pg 10                    ***********************
 **********************************************************************************************/

void MCCPTA()
{

  cout << endl << "************  Multiclass Continuous Perceptron Training Algorithm  ************";
  cout << endl << "************                   Lecture 7, Pg 10                    ************" << endl;

  const int J = 3;                   // no. of elements in Y, W vectors
  const int P = 3;                     // no. of training patterns
  const int R = 3;                     // no. of classes
  const double Emax = 0.01;               // maximum error
  
  typedef CWTSVector<J> VectorY;

  int p, i, g, cycleno, j;
  double threshold = -1.0;
  double lambda = 1;
  double nu = 0.1;
  double e, o, net;
  double d;

  /**** Results: ***************************************************************
   **** To implement this algorithm, this program uses the example given in
   **** the notes for the Multiclass Discrete Perceptron. The user may also
   **** input other values for the vectors than the ones used by default.
   **** 
   *****************************************************************************/

  VectorY y[P];                        // array of input vectors, i.e. the input patterns.
  for (p = 0; p < P; p++)
      y[p].Fill(threshold);
  y[0][0] = 10; y[0][1] = 2;
  y[1][0] = 2;  y[1][1] = -5;
  y[2][0] = -5; y[2][1] = 5;

  VectorY w[P];                        // weight vector W
  // The weight vector used in the Multiclass Discrete Percepton example
  /*  w[0][0] = 1; w[0][1] = -2; w[0][2] = 0;
      w[1][0] = 0; w[1][1] = -1; w[1][2] = 2;
      w[2][0] = 1; w[2][1] = 3;  w[2][2] = -1;*/

  // Initializing W with small random values
  w[0][0] = 0.4; w[0][1] = -0.2; w[0][2] = 0.1;
  w[1][0] = 0.5; w[1][1] = -0.1; w[1][2] = 0.2;
  w[2][0] = 0.1; w[2][1] = 0.3;  w[2][2] = -0.1;

  char choice, c;
  cout << endl << "No. of Input Patterns P (" << P << "): "; 
  cout << endl << "No. of Classes R (" << R << "): "; 
  cout << endl << "Threshold (" << threshold << "): ";
  cout << endl << "nu (" << nu << "): ";
  cout << endl << "lambda (" << lambda << "): ";
  for (i = 0; i < P; i++)
    {
      cout << endl << "Input Vector Y" << i+1; iprint(": ", y[i]);
      cout << endl << "Weight Vector W" << i+1 << ": "; iprint(w[i]);
    }
  
  cout << endl << "Do you want to use the default values? (y/n): ";
  cin >> choice;
  if ((choice == 'n') || (choice == 'N'))
    {
      cout << endl << "Preparing to accept values from the user...";
      cout << endl << "The original values are shown in parenthesis." << endl;
      cout << endl << "Threshold (" << threshold << "): "; cin >> threshold;
      cout << "nu (" << nu << "): "; cin >> nu;
      cout << "lambda (" << lambda << "): "; cin >> lambda;
      for (i = 0; i < P; i++)
	for (j = 0; j < R; j++)
	  {
	    cout << "Enter value for Y" << i+1 << " " << j << " (" << y[i][j] << "): ";
	    cin >> y[i][j];
	    cout << "Enter value for w" << i+1 << " " << j << " (" << w[i][j] << "): ";
	    cin >> w[i][j];
	  }
    }

  cycleno = 0;
  e = Emax+1;
  while (e > Emax)
    {
      e = 0;
      cycleno++;
      cout << endl << "Learning Cycle No.: " << cycleno;
      // for each class
      for (i = 0; i < R; i++)
	{
	  cout << endl << "\t## Class " << i+1 << " ##";
	  // for each pattern
	  for (g = 0; g < P; g++)
	    {
	      // calculate output
	      net = w[g] * y[i];
	      o = bipolar(lambda, net);
	      cout << endl << "\to" << g+1 << ": " << o;
	      // select the value of "d", i.e. the expected output
	      if (i == g)
		d = 1;
	      else
		d = -1;
	      cout << "   d = " << d;
	      // update the weight vector
	      w[g] = w[g] + ( y[i] * ( 0.5 * nu * (d - o) * (1 - pow(o,2)) ) );
	      e = e + (0.5 * pow((d - o), 2));
	      cout << "   e = " << e;
	      cout << endl << w[g];
	    }
	}
    }

  cout << endl << "Training terminated: further weight adjustment unrequired." << endl;
  for (i = 0; i < P; i++)
    {
      cout << "w" << i+1 << ": ";
      iprint(w[i], "\n");
    }

  // TESTING the NN.
  for (g = 0; g < P; g++)
    {
      net = w[g] * y[i];
      o = bipolar(lambda, net);
      if (i == g)
	d = 1;
      else
	d = -1;
      cout << endl << "Input pattern Y" << g+1; iprint(" ", y[g]);
      cout << endl << "o" << g+1 << ": " << o;
      cout << endl << "Output = " << rint(o) << "  d = " << d << endl;
    }

}
