/**********************************************************************************************
 *****************      FINAL PROJECT IN NEURAL NETWORKS -- CS 5388     ***********************
 *****************                 by: Karan M. Gupta                   ***********************
 **********************************************************************************************
 *****************     Single Continous Perceptron Training Algorithm   ***********************
 *****************                  Lecture 6, Pg 24                    ***********************
 **********************************************************************************************/

void SCPTA()
{

  cout << endl << "************  Single Continuous Perceptron Training Algorithm  ************";
  cout << endl << "************                 Lecture 6, Pg 24                  ************" << endl;

  const int J = 2;                  // no. of elements in Y, W vectors
  const int P = 4;                    // no. of (x,d) training pairs

  typedef CWTSVector<J> VectorY;

  int i, g, cycleno;
  double net, e, i0;
  double threshold = 1.0;
  double nu = 0.1;
  double lambda = 1;

  /**** Results: *****************************************************************
   ****	Running the algorithm with same values for the input and weight vectors
   ****	as the Single Discrete algorithm.
   ****	The w[0] and w[1] values are diverging in opposite directions.
   ****	1. Emax = 0.01 gives approximately W[3.94, -0.88]
   ****	1. Emax = 0.05 gives approximately W[2.82, -0.5]
   ****	1. Emax = 0.09 gives approximately W[2.45, -0.27]
   **** 1. Emax = 0.1  gives approximately W[2.40, -0.223 ]
   ****	1. Emax = 1    gives approximately W[1.32, 1.24]
   *******************************************************************************
   **** After the learning process has terminated, the NN is tested. Its output
   **** for each of the input vectors is compared with the teacher input, "d".
   **** Both the output and the rounded output are displayed. It is observed that 
   **** the smaller the value of Emax, the better the output values.
   *******************************************************************************/
  double Emax = 0.01;

  VectorY y[P];                          // array of input pattern vectors
  for (i = 0; i < P; i++)
      y[i].Fill(threshold);
  y[0][0] = 1;      // y[0][1] = threshold
  y[1][0] = -0.5;   // y[1][1] = threshold
  y[2][0] = 3;      // y[2][1] = threshold
  y[3][0] = -2;     // y[3][1] = threshold

  VectorY w;                            // weight vector W
  w[0] = -2.5; w[1] = 1.75;

  double d[] = {1, -1, 1, -1};          // Teacher Inputs

  char choice, c;
  cout << endl << "No. of Input Patterns P (" << P << "): "; 
  cout << endl << "Threshold (" << threshold << "): ";
  cout << endl << "nu (" << nu << "): ";
  cout << endl << "lambda (" << lambda << "): ";
  for (i = 0; i < P; i++)
    {
      cout << endl << "Input Vector Y" << i+1; iprint(": ", y[i]);
      cout << "  d" << i+1 << ": " << d[i];
    }
  cout << endl << "Weight Vector W: "; iprint(w);
  
  cout << endl << "Do you want to use the default values? (y/n): ";
  cin >> choice;
  if ((choice == 'n') || (choice == 'N'))
    {
      cout << endl << "Preparing to accept values from the user...";
      cout << endl << "The original values are shown in parenthesis." << endl;
      cout << endl << "Threshold (" << threshold << "): "; cin >> threshold;
      cout << "nu (" << nu << "): "; cin >> nu;
      cout << "lambda (" << lambda << "): "; cin >> lambda;
      for (i = 0; i < P; i++)
	{
	  cout << "Enter value for Y" << i+1 << " 0 (" << y[i][0] << "): ";
	  cin >> y[i][0];
	  cout << "Enter value for d" << i+1 << " (" << d[i] << "): ";
	  cin >> d[i];
	}
      cout << "Enter value for w0 (" << w[0] << "): "; cin >> w[0];
      cout << "Enter value for w1 (" << w[1] << "): "; cin >> w[1];
    }

  cycleno = 0;
  e = Emax + 1;
  while(e > Emax)
    {
      e = 0;      
      cycleno++;
      cout << endl << "****** Learning Cycle No.: " << cycleno << " ******" << endl;
      // for each pattern
      for (g = 0; g < P; g++)
	{
	  // calculate output
	  net = w * y[g];
	  i0 = bipolar(lambda, net);
	  // calculate new values for the weight vector
	  w = w + y[g] * ( 0.5 * nu * (d[g] - i0) * (1 - pow(i0,2)) );
	  e = e + ( 0.5 * pow( (d[g] - i0), 2) );
	  cout << endl << "e = " << e;
	  cout << endl << "w = " << endl << w;	  
	}
    }
  
  // Terminate Training.
  cout << endl << "Training terminated.";
  cout << endl << "e = " << e;
  cout << endl << "Total Learning Cycles = " << cycleno;
  iprint("\nw = ", w, "\n");

  // Test if the NN has learnt or not.
  cout << endl << endl << "**************     TESTING     ****************";
  for (g = 0; g < P; g++)
    {
      net = w * y[g];
      i0 = bipolar(lambda, net);
      cout << endl << "Input Vector: "; iprint(y[g]);
      cout << endl << "output = " << i0;
      cout << endl << "o = " << rint(i0) << "  d = " << d[g] << endl;
    }
  cout << endl;
	
}
